/*
        Copyright (c) 2012, MICEX-RTS. All rights reserved.

        Plaza-2 Client Gate API Usage Sample.
        Replication DataStream Client.

        All the software and documentation included in this and any
        other MICEX-RTS CGate Releasese is copyrighted by MICEX-RTS.

        Redistribution and use in source and binary forms, with or without
        modification, are permitted only by the terms of a valid
        software license agreement with MICEX-RTS.

        THIS SOFTWARE IS PROVIDED "AS IS" AND MICEX-RTS DISCLAIMS ALL WARRANTIES
        EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION, ANY IMPLIED WARRANTIES OF
        NON-INFRINGEMENT, MERCHANTABILITY OR FITNESS FOR A PARTICULAR
        PURPOSE.  MICEX-RTS DOES NOT WARRANT THAT USE OF THE SOFTWARE WILL BE
        UNINTERRUPTED OR ERROR-FREE.  MICEX-RTS SHALL NOT, UNDER ANY CIRCUMSTANCES, BE
        LIABLE TO LICENSEE FOR LOST PROFITS, CONSEQUENTIAL, INCIDENTAL, SPECIAL OR
        INDIRECT DAMAGES ARISING OUT OF OR RELATED TO THIS AGREEMENT OR THE
        TRANSACTIONS CONTEMPLATED HEREUNDER, EVEN IF MICEX-RTS HAS BEEN APPRISED OF
        THE LIKELIHOOD OF SUCH DAMAGES.
*/

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <math.h>

#include <set>
#include <map>

#ifdef _WIN32
#include <windows.h>
#endif//_WIN32

#define CG_SHORT_NAMES
#include <cgate.h>

// helper macro to call CGate function
// and exit if there was error

#define CHECK_FAIL(f) \
	{ \
		uint32_t res = f; \
		if (res != CG_ERR_OK) \
		{ \
			log_info("Client gate error: %x", res); \
			exit(1); \
		} \
	} 

// helper macro to call CGate function
// and print warning if there was error

#define WARN_FAIL(f) \
	{ \
		uint32_t res = f; \
		if (res != CG_ERR_OK) \
			log_info("Error: %x", res); \
	} 


// helper variable used as a notification flag to exit
static volatile int bExit = 0;

#ifdef _WIN32
BOOL WINAPI InterruptHandler(DWORD reason)
{
	log_info("----BREAK----");
	bExit = 1;
	return 1;
}
#endif//_WIN32

enum Dir
{
	DirBuy,
	DirSell
};

struct Order
{
	uint64_t id;
	double price;
	int vol;
	Order(uint64_t id_, double price_, int vol_):id(id_),price(price_),vol(vol_){}
	bool operator = (const Order& other) const
	{
		return id == other.id;
	}
};

struct OrderComparatorBuy
{
	bool operator() (const Order* o1, const Order* o2) const
	{
		return (o1->price == o2->price) ? o1->id < o2->id : o1->price > o2->price;
	}
};

struct OrderComparatorSell
{
	bool operator() (const Order* o1, const Order* o2) const
	{
		return (o1->price == o2->price) ? o1->id < o2->id : o1->price < o2->price;
	}
};

template <class TComparator>
class Orders
{
	typedef std::map<int64_t, Order*> TOrdersById;
	typedef std::set<Order*, TComparator> TOrdersByPrice;
	TOrdersByPrice _orders_by_price;
	TOrdersById _orders_by_id;

	Order* _best;
	TComparator _compare;

public:
	Orders():_best(0){}

	void add(int64_t id, double price, int vol)
	{
		Order* order = new Order(id, price, vol);
		std::pair<typename TOrdersByPrice::iterator,bool> pairPriceOrder = _orders_by_price.insert(order);
		typename TOrdersByPrice::iterator itPriceOrder = pairPriceOrder.first;
		_orders_by_id[id] = order;
		if (_best)
		{
			_best = _compare(_best, order) ? _best : order;
		}
		else
			_best = order;
	}

	void modify(int64_t id, int vol)
	{
		if (vol == 0) {
			remove(id);
		}
		else
		{
			TOrdersById::iterator itOrder = _orders_by_id.find(id);
			if (itOrder == _orders_by_id.end())
			{
				fprintf(stderr, "\nGot modify operator for non-existing order id=%lld, vol=%d\n", id, vol);
				return;
			}
			Order* order = itOrder->second;
			order->vol = vol;
		}
	}

	void remove(int64_t id)
	{
		TOrdersById::iterator itIdOrder = _orders_by_id.find(id);
		if (itIdOrder == _orders_by_id.end())
		{
			fprintf(stderr, "\nGot remove operator for non-existing order id=%lld\n", id);
			return;
		}
		Order* order = itIdOrder->second;

		Order ordtemp(id, order->price, order->vol);
		typename TOrdersByPrice::iterator itPriceOrder = _orders_by_price.find(&ordtemp);
		assert (itPriceOrder != _orders_by_price.end());

		_orders_by_id.erase(itIdOrder);
		_orders_by_price.erase(itPriceOrder);

		if (_best == order)
			_best = _orders_by_price.empty() ? 0 : *_orders_by_price.begin();

		delete order;
	}

	Order* getBest() { return _best; }
};

class OrderBook
{
	int32_t _isin_id;
	Orders<OrderComparatorBuy> _orders_buy;
	Orders<OrderComparatorSell> _orders_sell;
public:
	OrderBook(int32_t isin_id):_isin_id(isin_id){}

	void add(Dir dir, int64_t id, double price, int vol)
	{
		if (dir == DirBuy)
			_orders_buy.add(id, price, vol);
		else
			_orders_sell.add(id, price, vol);
	}

	void modify(Dir dir, int64_t id, int vol)
	{
		if (dir == DirBuy)
			_orders_buy.modify(id, vol);
		else
			_orders_sell.modify(id, vol);
	}

	void remove(Dir dir, uint64_t id)
	{
		if (dir == DirBuy)
			_orders_buy.remove(id);
		else
			_orders_sell.remove(id);
	}

	Order* getBestBuy()
	{
		return _orders_buy.getBest();
	}

	Order* getBestSell()
	{
		return _orders_sell.getBest();
	}
};


#define MSG_ORDERS "orders"
#define MSG_ORDERS_LOG "orders_log"

#define FIELD_REPLACT "replAct"
#define FIELD_ACTION "action"
#define FIELD_ORD_ID "id_ord"
#define FIELD_ISIN_ID "isin_id"
#define FIELD_PRICE "price"
#define FIELD_VOL "amount_rest"
#define FIELD_STATUS "status"
#define FIELD_DIR "dir"

#define ACTION_CANCEL 0
#define ACTION_ADD 1
#define ACTION_EXEC 2

#define STATUS_NONSYSTEM 0x04

class OrderBookApp
{
	struct StreamData
	{
		size_t _msg_index;
		size_t _offset_replact;
		size_t _offset_action;
		size_t _offset_id;
		size_t _offset_isin_id;
		size_t _offset_price;
		size_t _offset_vol;
		size_t _offset_status;
		size_t _offset_dir;
	};
	StreamData _data_orders;
	StreamData _data_orders_log;
	bool _scheme_ready;

	Order* _best_buy;
	Order* _best_sell;

	typedef std::map<int32_t, OrderBook> TOrderBooks;
	TOrderBooks _books;

	uint32_t _target_isin_id;

	static CG_RESULT MessageCallback(cg_conn_t* conn, cg_listener_t* listener, struct cg_msg_t* msg, void* data)
	{
		OrderBookApp* _this = (OrderBookApp*)data;
		switch (msg->type)
		{
		case CG_MSG_STREAM_DATA: 
			{			
				struct cg_msg_streamdata_t* replmsg = (struct cg_msg_streamdata_t* )msg;

				StreamData* info = 0;
				if (replmsg->msg_index == _this->_data_orders._msg_index)
					info = &_this->_data_orders;
				else
				if (replmsg->msg_index == _this->_data_orders_log._msg_index)
					info = &_this->_data_orders_log;
				else
					return CG_ERR_OK;

				if (data)
				{
					char *data = (char *)replmsg->data;
					int32_t status = *(int32_t*)(data + info->_offset_status);

					if (status & STATUS_NONSYSTEM) // ignore non-system orders since we are not interested in them
						break;
					
					int64_t replact = *(int64_t*)(data + info->_offset_replact);
					int8_t action = *(data + info->_offset_action);
					if (replmsg->msg_index == _this->_data_orders._msg_index)
						action = 1;
					int64_t ord_id = *(int64_t*)(data + info->_offset_id);
					int32_t isin_id = *(int32_t*)(data + info->_offset_isin_id);
					int8_t dir = *(data + info->_offset_dir);
					
					TOrderBooks::iterator itBook = _this->_books.find(isin_id);
					if (itBook == _this->_books.end())
					{
						std::pair<TOrderBooks::iterator, bool> pairBook = _this->_books.insert(std::pair<int32_t, OrderBook>(isin_id, OrderBook(isin_id)));
						itBook = pairBook.first;
					}
					OrderBook& book = itBook->second;

					Dir orddir = dir == 1 ? DirBuy : DirSell;

					switch (action)
					{
					case ACTION_ADD:
						{
							int32_t vol = *(int32_t*)(data + info->_offset_vol);
							char* priceptr = data + info->_offset_price;
							int64_t priceint = 0;
							int8_t pricescale = 0;
							WARN_FAIL(cg_bcd_get(priceptr, &priceint, &pricescale));
							double price = ((double)priceint)/(pow(10.0, pricescale));

							book.add(orddir, ord_id, price, vol);
							break;
						}
					case ACTION_EXEC:
						{
							int32_t vol = *(int32_t*)(data + info->_offset_vol);

							book.modify(orddir, ord_id, vol);
							break;
						}
					case ACTION_CANCEL:
						{
							book.remove(orddir, ord_id);
							break;
						}
					default:
						fprintf(stderr, "\nUnknown action %d for order %I64d\n", action, ord_id);
					}


					if (isin_id == _this->_target_isin_id)
					{
						Order* bestbuy = book.getBestBuy();
						Order* bestsell = book.getBestSell();

						bool doUpdate = false;

						if (_this->_best_buy != bestbuy)
							{ _this->_best_buy = bestbuy; doUpdate = true; }
						if (_this->_best_sell != bestsell)
							{ _this->_best_sell = bestsell; doUpdate = true; }

						if (doUpdate)
						{

							printf("\rbuy %d (%f) - (%f) %d sell            ", 
								_this->_best_buy ? _this->_best_buy->vol : 0,
								_this->_best_buy ? _this->_best_buy->price : 0,
								_this->_best_sell ? _this->_best_sell->price : 0,
								_this->_best_sell ? _this->_best_sell->vol : 0);
								
						}

					}
				}
				break;
			}
		case CG_MSG_OPEN:
			{
				cg_scheme_desc_t* schemedesc = 0;
				WARN_FAIL(cg_lsn_getscheme(listener, &schemedesc));
				assert(schemedesc);

				cg_message_desc_t* msgdesc = schemedesc->messages;
				
				for (size_t index = 0; msgdesc; index ++, msgdesc = msgdesc->next)
				{
					cg_field_desc_t* fielddesc = msgdesc->fields;
					StreamData* data = 0;
					if (strcmp(msgdesc->name, MSG_ORDERS) == 0)
						data = &_this->_data_orders;
					else
					if (strcmp(msgdesc->name, MSG_ORDERS_LOG) == 0)
						data = &_this->_data_orders_log;
					else
						continue;

					data->_msg_index = index;

					while (fielddesc)
					{
						if (strcmp(fielddesc->name, FIELD_REPLACT) == 0)
						{
							data->_offset_replact = fielddesc->offset;
							assert (fielddesc->size == sizeof(int64_t));
						} else
						if (strcmp(fielddesc->name, FIELD_ACTION) == 0)
						{
							data->_offset_action = fielddesc->offset;
							assert (fielddesc->size == sizeof(int8_t));
						} else
						if (strcmp(fielddesc->name, FIELD_ORD_ID) == 0)
						{
							data->_offset_id = fielddesc->offset;
							assert (fielddesc->size == sizeof(int64_t));
						} else
						if (strcmp(fielddesc->name, FIELD_ISIN_ID) == 0)
						{
							data->_offset_isin_id = fielddesc->offset;
							assert (fielddesc->size == sizeof(int32_t));
						} else
						if (strcmp(fielddesc->name, FIELD_PRICE) == 0)
						{
							data->_offset_price = fielddesc->offset;
						} else
						if (strcmp(fielddesc->name, FIELD_VOL) == 0)
						{
							data->_offset_vol = fielddesc->offset;
							assert (fielddesc->size == sizeof(int32_t));
						} else
						if (strcmp(fielddesc->name, FIELD_STATUS) == 0)
						{
							data->_offset_status = fielddesc->offset;
							assert (fielddesc->size == sizeof(int32_t));
						} else
						if (strcmp(fielddesc->name, FIELD_DIR) == 0)
						{
							data->_offset_dir = fielddesc->offset;
							assert (fielddesc->size == sizeof(int8_t));
						}

						fielddesc = fielddesc->next;
					}
				}

				_this->_scheme_ready = true;
				break;
			}
		}

		return CG_ERR_OK;
	}

	cg_conn_t* _conn;
	cg_listener_t *_lsn_orders;
public:
	OrderBookApp(cg_conn_t* conn, const char* settings, uint32_t target_isin_id):
	  _conn(conn),
	  _scheme_ready(false),
	  _lsn_orders(0),
	  _best_buy(0),
	  _best_sell(0),
	  _target_isin_id(target_isin_id)
	{
		if (cg_lsn_new(_conn, settings, &MessageCallback, this, &_lsn_orders) != CG_ERR_OK)
			throw std::exception();
	}

	~OrderBookApp()
	{
		if (_lsn_orders)
			cg_lsn_destroy(_lsn_orders);
	}

	void run()
	{
		uint32_t state;
		WARN_FAIL(cg_lsn_getstate(_lsn_orders, &state));
		
		while (!bExit) // check global interrupt flag
		{
			uint32_t state;
			WARN_FAIL(cg_conn_getstate(_conn, &state));

			if (state == CG_STATE_ERROR)
			{
				WARN_FAIL(cg_conn_close(_conn));
			}
			else
			if (state == CG_STATE_CLOSED)
			{
				WARN_FAIL(cg_conn_open(_conn, 0));
			}
			else
			if (state = CG_STATE_ACTIVE)
			{
				uint32_t result = cg_conn_process(_conn, 0, 0);
				if (result != CG_ERR_OK && result != CG_ERR_TIMEOUT) 
				{
					log_info("Warning: connection state request failed: %X", result);
				}
			
				WARN_FAIL(cg_lsn_getstate(_lsn_orders, &state));

				switch (state)
				{
				case CG_STATE_CLOSED:
						WARN_FAIL(cg_lsn_open(_lsn_orders, 0));
						break;
				case CG_STATE_ERROR:
						WARN_FAIL(cg_lsn_close(_lsn_orders));
						break;						
				}
			}
		}
	}
};

int main(int argc, char* argv[])
{
	// connection pointer
	conn_t* conn = 0;
	// listener pointer
	listener_t* listener = 0;

	// connection init string
	const char* conn_str = "p2tcp://127.0.0.1:4001;app_name=test_ordbook";

#ifdef _WIN32
	SetConsoleCtrlHandler(InterruptHandler, 1);
#endif//_WIN32

	if (argc < 2) 
	{
		printf("Usage:\n\tordbook ISIN_ID\n\nwhere\n\tISIN_ID is unique instrument ID (e.g. 167566)\n");
		return 2;
	}

	uint32_t isin_id = atoi(argv[1]);

	// Environment is initialized using provided INI-file
	CHECK_FAIL(env_open("ini=ordbook.ini;key=11111111"));
	// Connection is initialized using connection init string
	CHECK_FAIL(conn_new(conn_str, &conn));

	try
	{
		const char* ordbook_str = "p2ordbook://FORTS_FUTTRADE_REPL;snapshot=FORTS_FUTORDERBOOK_REPL";
		OrderBookApp app(conn, ordbook_str, isin_id);
		app.run();
	}
	catch (std::exception& e)
	{
		fprintf(stderr, "Exception: %s", e.what());
	}

	// destroy connection
	if (conn != NULL)
		conn_destroy(conn);


	// close environment
	CHECK_FAIL(env_close());
	return 0;
}
